package telemetry

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"time"

	"github.com/gofrs/uuid"
	"github.com/sirupsen/logrus"
	"go.flipt.io/flipt/config"
	"go.flipt.io/flipt/internal/info"
	"gopkg.in/segmentio/analytics-go.v3"
)

const (
	filename = "telemetry.json"
	version  = "1.0"
	event    = "flipt.ping"
)

type ping struct {
	Version string `json:"version"`
	UUID    string `json:"uuid"`
	Flipt   flipt  `json:"flipt"`
}

type flipt struct {
	Version string `json:"version"`
}

type state struct {
	Version       string `json:"version"`
	UUID          string `json:"uuid"`
	LastTimestamp string `json:"lastTimestamp"`
}

type Reporter struct {
	cfg    config.Config
	logger logrus.FieldLogger
	client analytics.Client
}

func NewReporter(cfg config.Config, logger logrus.FieldLogger, analytics analytics.Client) *Reporter {
	return &Reporter{
		cfg:    cfg,
		logger: logger,
		client: analytics,
	}
}

type file interface {
	io.ReadWriteSeeker
	Truncate(int64) error
}

// Report sends a ping event to the analytics service.
func (r *Reporter) Report(ctx context.Context, info info.Flipt) (err error) {
	f, err := os.OpenFile(filepath.Join(r.cfg.Meta.StateDirectory, filename), os.O_RDWR|os.O_CREATE, 0644)
	if err != nil {
		return fmt.Errorf("opening state file: %w", err)
	}
	defer f.Close()

	return r.report(ctx, info, f)
}

func (r *Reporter) Close() error {
	return r.client.Close()
}

// report sends a ping event to the analytics service.
// visible for testing
func (r *Reporter) report(_ context.Context, info info.Flipt, f file) error {
	if !r.cfg.Meta.TelemetryEnabled {
		return nil
	}

	var s state

	if err := json.NewDecoder(f).Decode(&s); err != nil && !errors.Is(err, io.EOF) {
		return fmt.Errorf("reading state: %w", err)
	}

	// if s is empty we need to create a new state
	if s.UUID == "" {
		s = newState()
		r.logger.Debug("initialized new state")
	} else {
		t, _ := time.Parse(time.RFC3339, s.LastTimestamp)
		r.logger.Debugf("last report was: %v ago", time.Since(t))
	}

	var (
		props = analytics.NewProperties()
		p     = ping{
			Version: s.Version,
			UUID:    s.UUID,
			Flipt: flipt{
				Version: info.Version,
			},
		}
	)

	// marshal as json first so we can get the correct case field names in the analytics service
	out, err := json.Marshal(p)
	if err != nil {
		return fmt.Errorf("marshaling ping: %w", err)
	}

	if err := json.Unmarshal(out, &props); err != nil {
		return fmt.Errorf("unmarshaling ping: %w", err)
	}

	if err := r.client.Enqueue(analytics.Track{
		AnonymousId: s.UUID,
		Event:       event,
		Properties:  props,
	}); err != nil {
		return fmt.Errorf("tracking ping: %w", err)
	}

	s.Version = version
	s.LastTimestamp = time.Now().UTC().Format(time.RFC3339)

	// reset the state file
	if err := f.Truncate(0); err != nil {
		return fmt.Errorf("truncating state file: %w", err)
	}
	if _, err := f.Seek(0, 0); err != nil {
		return fmt.Errorf("resetting state file: %w", err)
	}

	if err := json.NewEncoder(f).Encode(s); err != nil {
		return fmt.Errorf("writing state: %w", err)
	}

	return nil
}

func newState() state {
	var uid string

	u, err := uuid.NewV4()
	if err != nil {
		uid = "unknown"
	} else {
		uid = u.String()
	}

	return state{
		Version: version,
		UUID:    uid,
	}
}
